# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Entry points that run the content generation tasks. Typically use these entry
points for testing and for prepopulating build directories. These functions do
not run in common practice.
"""

import logging

logger = logging.getLogger('giza.operations.generate')

import argh

from giza.core.app import BuildApp
from giza.core.app import build_app_context
from giza.config.helper import fetch_config
from giza.tools.files import rm_rf

from giza.content.assets import assets_tasks, assets_clean
from giza.content.images import image_tasks, image_clean
from giza.content.intersphinx import intersphinx_tasks, intersphinx_clean
from giza.content.param import api_tasks, api_clean
from giza.content.links import create_manual_symlink
from giza.content.table import table_tasks, table_clean
from giza.content.robots import robots_txt_tasks
from giza.content.redirects import make_redirect, redirect_tasks
from giza.content.tocs.tasks import toc_tasks
from giza.content.examples.tasks import example_tasks
from giza.content.steps.tasks import step_tasks, step_clean
from giza.content.options.tasks import option_tasks, option_clean
from giza.content.release.tasks import release_tasks, release_clean

from giza.content.primer import primer_migration_tasks
from giza.content.primer import clean as primer_clean

from giza.operations.sphinx_cmds import build_content_generation_tasks
from giza.content.source import source_tasks
from giza.config.sphinx_config import render_sconf
from giza.content.dependencies import refresh_dependency_tasks

import giza.content.apiargs.migration

@argh.arg('--edition', '-e')
@argh.expects_obj
def toc(args):
    c = fetch_config(args)

    with build_app_context(c) as app:
        for task in toc_tasks(c):
            app.add(task)

@argh.arg('--edition', '-e')
@argh.arg('--clean', '-c', default=False, action="store_true", dest="clean_generated")
@argh.expects_obj
def steps(args):
    c = fetch_config(args)

    with build_app_context(c) as app:
        if c.runstate.clean_generated is True:
            step_clean(c, app)
        else:
            app.extend_queue(step_tasks(c))

@argh.arg('--clean', '-c', default=False, action="store_true", dest="clean_generated")
@argh.expects_obj
def options(args):
    c = fetch_config(args)

    if c.runstate.clean_generated is True:
        option_clean(c)
    else:
        with build_app_context(c) as app:
            app.extend_queue(option_tasks(c))

@argh.arg('--clean', '-c', default=False, action="store_true", dest="clean_generated")
@argh.expects_obj
def api(args):
    c = fetch_config(args)

    with build_app_context(c) as app:
        if c.runstate.clean_generated is True:
            api_clean(c, app)
        else:
            api_tasks(c, app)

@argh.arg('--clean', '-c', default=False, action="store_true", dest="clean_generated")
@argh.expects_obj
def assets(args):
    c = fetch_config(args)

    with build_app_context(c) as app:
        if c.runstate.clean_generated is True:
            assets_clean(c, app)
        else:
            assets_tasks(c, app)

@argh.arg('--clean', '-c', default=False, action="store_true", dest="clean_generated")
@argh.expects_obj
def images(args):
    c = fetch_config(args)
    app = BuildApp(c)

    with build_app_context(c) as app:
        if c.runstate.clean_generated is True:
            image_clean(c, app)
        else:
            image_tasks(c, app)

@argh.arg('--clean', '-c', default=False, action="store_true", dest="clean_generated")
@argh.expects_obj
def intersphinx(args):
    c = fetch_config(args)

    with build_app_context(c) as app:
        if c.runstate.clean_generated is True:
            intersphinx_clean(c, app)
        else:
            intersphinx_tasks(c, app)

@argh.arg('--clean', '-c', default=False, action="store_true", dest="clean_generated")
@argh.expects_obj
def primer(args):
    c = fetch_config(args)

    with build_app_context(c) as app:
        if c.runstate.clean_generated is True:
            primer_clean(c, app)
        else:
            primer_migration_tasks(c, app)

@argh.arg('--clean', '-c', default=False, action="store_true", dest="clean_generated")
@argh.expects_obj
def release(args):
    c = fetch_config(args)

    with build_app_context(c) as app:
        if c.runstate.clean_generated is True:
            release_clean(c, app)
        else:
            app.extend_queue(release_tasks(c))

@argh.arg('--clean', '-c', default=False, action="store_true", dest="clean_generated")
@argh.expects_obj
def tables(args):
    c = fetch_config(args)

    with build_app_context(c) as app:
        if c.runstate.clean_generated is True:
            table_clean(c, app)
        else:
            table_tasks(c, app)

@argh.expects_obj
def examples(args):
    c = fetch_config(args)

    with build_app_context(c) as app:
        app.extend_queue(example_tasks(c))

@argh.arg('--edition', '-e')
@argh.expects_obj
def robots(args):
    c = fetch_config(args)

    with build_app_context(c) as app:
        app.pool = 'serial'
        robots_txt_tasks(c, app)

@argh.arg('--edition', '-e')
@argh.arg('--print', '-p', default=False, action='store_true', dest='dry_run')
@argh.expects_obj
def redirects(args):
    c = fetch_config(args)

    if args.dry_run is True:
        print(''.join(make_redirect(c)))
    else:
        with build_app_context(c) as app:
            redirect_tasks(c, app)

from giza.content.primer import primer_migration_tasks
from giza.content.assets import assets_tasks


@argh.arg('--edition', '-e')
@argh.arg('--language', '-l')
@argh.expects_obj
def source(args):
    conf = fetch_config(args)

    sconf = render_sconf(args.edition, 'html', args.language, conf)
    with build_app_context(conf) as app:
        with app.context() as pre_app:
            assets_tasks(conf, pre_app.add('app'))
            primer_migration_tasks(conf, pre_app)

        with app.context() as source_app:
            source_tasks(conf, sconf, source_app)

        for content, func in build_config.system.content.task_generators:
            app.add(Task(job=func,
                         args=[build_config],
                         target=True))

        content_generator_tasks = app.run()
        app.reset()

        for group in content_generator_tasks:
            app.extend_queue(group)

        build_content_generation_tasks(conf, app.add('app'))
        refresh_dependency_tasks(conf, app.add('app'))

@argh.expects_obj
def migration(args):
    conf = fetch_config(args)

    with build_app_context(conf) as app:
        name_changes = giza.content.apiargs.migration.task(task='source', conf=conf)

        for task in giza.content.apiargs.migration.file_munge_tasks(name_changes, 'source', conf):
            app.add(task)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import logging

logger = logging.getLogger('giza.content.extract.tasks')

from giza.tools.files import safe_create_directory
from giza.tools.transformation import append_to_file, prepend_to_file
from giza.content.extract.inheritance import ExtractDataCache
from giza.content.extract.views import render_extracts, get_include_statement
from giza.config.content import new_content_type
from giza.core.task import Task

def register_extracts(conf):
    conf.system.content.add(name='extracts', definition=new_content_type(name='extracts', task_generator=extract_tasks, conf=conf))

def write_extract_file(extract, fn):
    content = render_extracts(extract)
    content.write(fn)
    logger.info('wrote extract file: ' + fn)

def extract_tasks(conf):
    extracts = ExtractDataCache(conf.system.content.extracts.sources, conf)
    extracts.create_output_dir()

    tasks = []
    for dep_fn, extract in extracts.content_iter():
        t = Task(job=write_extract_file,
                 args=(extract, extract.target),
                 description="generating extract file: " + extract.target,
                 target=extract.target,
                 dependency=dep_fn)
        tasks.append(t)

        include_statement = get_include_statement(extract.target_project_path)

        for verb, adjc, files in [ (prepend_to_file, 'prepend', extract.prepend),
                                  (append_to_file, 'append', extract.append) ]:
            for fn in files:
                t = Task(job=verb,
                         args=(fn, include_statement),
                         target=fn,
                         dependency=None,
                         description="{0} extract include for '{0}' to '{1}'".format(adjc, extract.target, fn))
                tasks.append(t)

    logger.info("added tasks for {0} extract generation tasks".format(len(tasks)))

    return tasks



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import datetime
import itertools
import logging
import os
import tarfile
import tempfile

import argh
from sphinx.application import Sphinx, ENV_PICKLE_FILENAME
from sphinx.builders.html import get_stable_hash

from giza.config.sphinx_config import avalible_sphinx_builders, resolve_builder_path
from giza.operations.packaging import fetch_package
from giza.operations.sphinx_cmds import get_sphinx_build_configuration
from giza.tools.files import cd, safe_create_directory, FileNotFoundError
from giza.tools.strings import hyph_concat
from giza.core.app import build_app_context

logger = logging.getLogger('giza.operations.build_env')

#################### Helpers ####################

def is_git_dir(path):
    git_dir = ''.join([ os.path.sep, '.git', os.path.sep])
    if git_dir in path:
        return True
    else:
        return False

def extract_package_at_root(path, conf):
    with cd(conf.paths.projectroot):
        with tarfile.open(path, "r:gz") as t:
            t.extractall()

def get_existing_builders(conf):
    return [ b
             for b in avalible_sphinx_builders()
             if os.path.isdir(os.path.join(conf.paths.projectroot, conf.paths.branch_output, b)) ]

def env_package_worker(args, conf):
    # used by the make interface
    package_build_env(args.builder, args.editions_to_build, args.languages_to_build, conf)

#################### Core Workers ####################

def package_build_env(builders, editions, languages, conf):
    arc_fn = hyph_concat('cache', conf.project.name, conf.git.branches.current, datetime.datetime.utcnow().strftime('%s'), conf.git.commit[:8]) + ".tar.gz"
    archive_path = os.path.join(conf.paths.buildarchive, arc_fn)
    safe_create_directory(conf.paths.buildarchive)

    existing_archives = os.listdir(conf.paths.buildarchive)

    for arc in existing_archives:
        if conf.git.commit[:8] in arc:
            logger.warning('archive "{0}" exists for current git hash, not recreating'.format(archive_path))
            return

    logger.debug("no archive for commit '{0}' continuing".format(conf.git.commit))

    with cd(conf.paths.projectroot):
        files_to_archive = set()

        for edition, language, builder in itertools.product(editions, languages, builders):
            rconf, sconf = get_sphinx_build_configuration(edition, language, builder, copy.deepcopy(conf.runstate))
            builder_dirname = resolve_builder_path(builder, edition, language, rconf)

            files_to_archive.add(rconf.paths.branch_source)
            files_to_archive.add(os.path.join(rconf.paths.branch_output, builder_dirname))
            files_to_archive.add(os.path.join(rconf.paths.branch_output, hyph_concat('doctrees', builder_dirname)))
            files_to_archive.add(rconf.system.dependency_cache_fn)

        files_to_archive = list(files_to_archive)
        logger.info('prepped build cache archive. writing file now.')

        for fn in files_to_archive:
            if not os.path.exists(fn):
                raise FileNotFoundError(fn)

        try:
            with tarfile.open(archive_path, 'w:gz') as t:
                for fn in files_to_archive:
                    t.add(fn, exclude=is_git_dir)
            logger.info("created build-cache archive: " + archive_path)
        except Exception as e:
            os.remove(archive_path)
            logger.critical("failed to create archive: " + archive_path)
            logger.error(e)

def fix_build_env(builder, conf):
    """
    Given a builder name and the conf object, this function fixes the build
    artifacts for the current build to prevent a full rebuild. Currently
    re-pickles the environment and dumps the ``.buildinfo`` file in the build
    directory with the correct hashes.
    """

    fn = os.path.join(conf.paths.projectroot, conf.paths.branch_output, builder, '.buildinfo')
    logger.info('updating cache for: ' + builder)

    if not os.path.isfile(fn):
        return

    doctree_dir = os.path.join(conf.paths.projectroot, conf.paths.branch_output, "doctrees-" + builder)

    sphinx_app = Sphinx(
        srcdir=os.path.join(conf.paths.projectroot, conf.paths.branch_output, "source"),
        confdir=conf.paths.projectroot,
        outdir=os.path.join(conf.paths.projectroot, conf.paths.branch_output, builder),
        doctreedir=doctree_dir,
        buildername=builder,
        status=tempfile.NamedTemporaryFile(),
        warning=tempfile.NamedTemporaryFile())

    sphinx_app.env.topickle(os.path.join(doctree_dir, ENV_PICKLE_FILENAME))

    with open(fn, 'r') as f:
        lns = f.readlines()
        tags_hash_ln = None
        for ln in lns:
            if ln.startswith('tags'):
                tags_hash_ln = ln
                break

        if tags_hash_ln == None:
            tags_hash_ln = 'tags: '  + get_stable_hash(sorted(sphinx_app.tags))

    with open(fn, 'w') as f:
        f.write('# Sphinx build info version 1')
        f.write('\n\n') ## current format requires an extra line here.
        f.write('config: ' + get_stable_hash(dict((name, sphinx_app.config[name])
                                                  for (name, desc) in sphinx_app.config.values.items()
                                                  if desc[1] == 'html')))
        f.write('\n')
        f.write(tags_hash_ln)
        f.write('\n')

#################### Task Creators ####################

def fix_build_env_tasks(builders, conf, app):
    for builder in builders:
        t = app.add('task')
        t.job = fix_build_env
        t.args = (builder, conf)
        t.target = True
        t.description = "fix up sphinx environment for builder '{0}'".format(builder)

#################### Entry Points ####################

@argh.arg('--edition', '-e', nargs='*', dest='editions_to_build')
@argh.arg('--language', '-l', nargs='*',dest='languages_to_build')
@argh.arg('--builder', '-b', nargs='*', default='html')
@argh.expects_obj
def package(args):
    conf = fetch_config(args)

    package_build_env(args.builder, args.editions_to_build, args.languages_to_build, conf)

@argh.arg('--path', '-p', default=None, dest='_path')
@argh.expects_obj
def extract(args):
    conf = fetch_config(args)

    with build_app_context(conf) as app:
        path = fetch_package(args._path, conf)
        extract_package_at_root(path, conf)

        builders = get_existing_builders(conf)
        fix_build_env_tasks(builders, conf, app)



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import logging
import os
import copy
import difflib

logger = logging.getLogger('giza.content.apiargs.migration')

from giza.tools.serialization import ingest_yaml_list, write_yaml, literal_str
from giza.tools.strings import hyph_concat
from giza.tools.transformation import munge_page
from giza.tools.files import expand_tree, safe_create_directory, rm_rf, verbose_remove
from giza.core.task import Task

def task(task, conf):
    if task == 'source':
        legacy_tables = expand_tree(os.path.join(conf.paths.projectroot, conf.paths.source, 'reference'), 'yaml')
        dirname = os.path.join(conf.paths.projectroot, conf.paths.includes, 'apiargs')
        safe_create_directory(dirname)
        offset = len(os.path.join(conf.paths.projectroot, conf.paths.source))
    elif task == 'branch':
        legacy_tables = expand_tree(os.path.join(conf.paths.projectroot, conf.paths.branch_source, 'reference'), 'yaml')
        safe_create_directory(conf.system.content.apiargs.output_dir)
        offset = len(os.path.join(conf.paths.projectroot, conf.paths.branch_source))
    else:
        logger.critical('cannot perform apiarg migration for: ' + str(task))
        return

    new_records = []
    new_basenames = []
    new_fns = []
    old_fns = []

    for fn in legacy_tables:
        new_data, new_fn = migrate_legacy_apiarg(task, fn, conf)
        if new_fn in new_fns:
            logger.error("duplicate: {0}, from: {1}".format(os.path.basename(new_fn), os.path.basename(fn)))
        else:
            new_basenames.append(new_fn[offset:])
            new_fns.append(new_fn)
            new_records.append(new_data)
            old_fns.append(fn)

    task_maker = zip(new_basenames, new_records, old_fns, new_fns)

    for basename, data, old_fn, new_fn in task_maker:
        write_yaml(data, new_fn)
        if task == 'source':
            verbose_remove(old_fn)

    new_sources = conf.system.content.apiargs.sources

    if len(new_sources) != len(legacy_tables) and len(legacy_tables) != len(new_fns):
        logger.critical('problem in apiargs table migration.')
    else:
        logger.info('legacy apiargs tables migrated successfully.')

    legacy_tables = [ fn[offset:] for fn in legacy_tables ]
    return zip(legacy_tables, new_basenames)

def migrate_legacy_apiarg(task, fn, conf, silent=False):
    legacy_data = ingest_yaml_list(fn)

    new_data, meta = transform_data(task, legacy_data, fn[len(os.path.join(conf.paths.projectroot, conf.paths.branch_output))+1:], silent, conf)

    old_base = os.path.basename(fn)
    if not old_base.startswith(meta['operation']):
        meta['operation'] = old_base[:-5].split('-', 1)[0]

    tag = old_base[:-5][len(meta['operation'])+1:]
    if tag.startswith('-'):
        tag = tag[1:]
    if tag == 'fields':
        tag = 'field'

    new_fn_base = hyph_concat('apiargs', meta['interface'], meta['operation'], tag)
    new_fn_base = new_fn_base + '.yaml'

    if task == 'source':
        new_fn = os.path.join(conf.paths.projectroot,
                              conf.paths.includes,
                              new_fn_base)
    elif task =='branch':
        new_fn = os.path.join(conf.paths.projectroot,
                              conf.paths.branch_includes,
                              new_fn_base)
    return new_data, new_fn

def transform_data(task, data, fn, silent, conf):
    output = []

    meta = {}
    for doc in data:
        if 'object' not in doc:
            if 'file' not in doc:
                print(doc)
                logger.error("error in: " + fn)
            else:
                operation = os.path.basename(fn).split('-')[0]
                if 'method' in fn:
                    meta['interface'] = 'method'
                    meta['operation'] = operation
                elif 'command' in fn:
                    meta['interface'] = 'dbcommand'
                    meta['operation'] = operation
                elif 'aggregation' in fn:
                    meta['interface'] = 'pipeline'
                    meta['operation'] = operation
                else:
                    logger.error("invalid document format in: " + fn)

                new_doc = copy.copy(meta)

                src_file = doc['file']
                if src_file.startswith('/'):
                    src_file = src_file[1:]

                if task == 'source':
                    src_file = os.path.join(conf.paths.projectroot, conf.paths.source, src_file)
                elif task == 'branch':
                    src_file = os.path.join(conf.paths.projectroot, conf.paths.branch_source, src_file)

                _, new_fn = migrate_legacy_apiarg(task, src_file, conf, True)

                new_doc.update({ 'source': { 'file': os.path.basename(new_fn), 'ref': doc['name']}})
                output.append(new_doc)

            continue

        if doc['object']['name'].endswith('()'):
            doc['object']['name'] = doc['object']['name'][:-2]

        if silent is False and ('interface' in meta and meta['interface'] != doc['object']['type']):
            logger.warning("interface type (cmd, method) values do not agree in: " + fn)
        meta['interface'] = doc['object']['type']

        if silent is False and ('operation' in meta and meta['operation'] != doc['object']['name']):
            logger.warning("calling operation names do not agree in: " + fn)
        meta['operation'] = doc['object']['name']

        if silent is False and ('arg_name' in meta and meta['arg_name'] != doc['field']['type']):
            logger.warning('argument types do not agree in: ' + fn)
        meta['arg_name'] = doc['field']['type']

        new_doc = copy.copy(meta)
        if 'optional' in doc['field']:
            new_doc['optional'] = doc['field']['optional']

        for field_name in ('name', 'type', 'position', 'description'):
            if field_name in doc:
                if field_name == 'description':
                    new_doc[field_name] = literal_str(doc[field_name])
                else:
                    new_doc[field_name] = doc[field_name]

        output.append(new_doc)

    return output, meta

def file_munge_tasks(name_changes, loc, conf):
    tasks = []

    if loc == 'branch':
        return tasks

    name_changes = [
        (old.replace('.yaml', '.rst'), new.replace('apiargs-', 'apiargs/').replace('.yaml', '.rst'))
        for old, new in name_changes
    ]

    if loc == 'source':
        ref_files = expand_tree(os.path.join(conf.paths.projectroot, conf.paths.branch_source, 'reference'), 'txt')
    elif loc == 'branch':
        ref_files = expand_tree(os.path.join(conf.paths.projectroot, conf.paths.source, 'reference'), 'txt')

    for old, new in name_changes:
        re_pattern = (re.compile(r'include:: ' + old), r'include:: ' + new)

        for ref_file in ref_files:
            t = Task(job=munge_page,
                     args=dict(fn=ref_file,
                               regex=re_pattern),
                     target=ref_file,
                     dependency=None,
                     description='munging reference page')
            tasks.append(t)

    return tasks



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging

logger = logging.getLogger('giza.content.release.tasks')

from giza.tools.files import expand_tree, verbose_remove, safe_create_directory
from giza.content.release.inheritance import ReleaseDataCache
from giza.content.release.views import render_releases
from giza.config.content import new_content_type
from giza.core.task import Task

def register_releases(conf):
    conf.system.content.add(name='releases', definition=new_content_type(name='release', task_generator=release_tasks, conf=conf))

def write_release_file(release, fn, conf):
    content = render_releases(release, conf)
    content.write(fn)
    logger.info('wrote release content: ' + fn)

def release_tasks(conf):
    rel = ReleaseDataCache(conf.system.content.releases.sources, conf)
    rel.create_output_dir()

    tasks = []
    for dep_fn, release in rel.content_iter():
        t = Task(job=write_release_file,
                 args=(release, release.target, conf),
                 description='generating release spec file: ' + release.target,
                 target=release.target,
                 dependency=dep_fn)
        tasks.append(t)

    logger.info("added tasks for {0} release generation tasks".format(len(tasks)))
    return tasks

def release_clean(conf, app):
    rel = ReleaseDataCache(conf.system.content.releases.sources, conf)

    for dep, release in rel.content_iter():
        task = app.add('task')
        task.target = True
        task.dependency = fn
        task.job = verbose_remove
        task.args = [release.target]
        task.description = 'removing {0}'.format(fn)
